@@ -1,12 +1,14 @@
 """Custom client for interacting with Monzo's API"""
-
-from logging import getLogger, DEBUG
+from __future__ import annotations
 
 from datetime import datetime, timedelta
+from logging import DEBUG, getLogger
+from typing import Dict, Generator, Literal, Optional, TypedDict, Union
+
 from requests import get, put
 
 from wg_utilities.clients._generic import OauthClient
-from wg_utilities.functions import user_data_dir, cleanse_string
+from wg_utilities.functions import cleanse_string, user_data_dir
 from wg_utilities.loggers import add_stream_handler
 
 LOGGER = getLogger(__name__)
@@ -16,19 +18,83 @@ add_stream_handler(LOGGER)
 DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
 
 
+class _MonzoAccountInfo(TypedDict):
+    account_number: str
+    balance: float
+    balance_including_flexible_savings: float
+    created: str
+    description: str
+    id: str
+    sort_code: str
+    spend_today: float
+    total_balance: float
+
+
+class _MonzoPotInfo(TypedDict):
+    available_for_bills: bool
+    balance: float
+    charity_id: str
+    cover_image_url: str
+    created: str
+    currency: str
+    current_account_id: str
+    deleted: bool
+    goal_amount: float
+    has_virtual_cards: bool
+    id: str
+    is_tax_pot: bool
+    isa_wrapper: str
+    locked: bool
+    name: str
+    product_id: str
+    round_up: bool
+    round_up_multiplier: float
+    style: str
+    type: str
+    updated: str
+
+
+class _BalanceVariablesInfo(TypedDict):
+    balance: int
+    balance_including_flexible_savings: int
+    currency: Literal["GBP"]
+    local_currency: str
+    local_exchange_rate: Union[int, float]
+    local_spend: Dict[str, Union[str, int]]
+    spend_today: int
+    total_balance: int
+
+
 class Account:
     """Class for managing individual bank accounts"""
 
-    def __init__(self, json, monzo_client=None, balance_update_threshold=15):
+    def __init__(
+        self,
+        json: _MonzoAccountInfo,
+        monzo_client: MonzoClient,
+        balance_update_threshold: int = 15,
+    ):
         self.json = json
         self._monzo_client = monzo_client
 
-        self._balance_variables = None
+        self._balance_variables: _BalanceVariablesInfo
 
         self.last_balance_update = datetime(1970, 1, 1)
         self.balance_update_threshold = balance_update_threshold
 
-    def _get_balance_property(self, var_name):
+    def _get_balance_property(
+        self,
+        var_name: Literal[
+            "balance",
+            "balance_including_flexible_savings",
+            "currency",
+            "local_currency",
+            "local_exchange_rate",
+            "local_spend",
+            "spend_today",
+            "total_balance",
+        ],
+    ) -> Union[str, float, Dict[str, Union[str, int]], None]:
         """Gets a value for a balance-specific property, updating the values if
          necessary (i.e. if they don't already exist). This also has a check to see if
          property is relevant for the given entity type and if not it just returns None
@@ -41,10 +107,10 @@ class Account:
         """
 
         if (
-            self._balance_variables is not None
+            hasattr(self, "_balance_variables")
             and var_name not in self._balance_variables
         ):
-            # assume it's not a valid value so there's no point polling the API again
+            # Assume it's not a valid value so there's no point polling the API again
             return None
 
         if self.last_balance_update <= (
@@ -54,24 +120,25 @@ class Account:
 
         return self._balance_variables[var_name]
 
-    def update_balance_variables(self):
+    def update_balance_variables(self) -> None:
         """Updates the balance-related instance attributes with the latest values from
         the API
         """
-
-        self._balance_variables = self._monzo_client.get_json_response(
+        # pylint: disable=line-too-long
+        self._balance_variables = self._monzo_client.get_json_response(  # type: ignore[assignment]
             "/balance", params={"account_id": self.id}
         )
 
         # convert values from pence to pounds
         for key, value in self._balance_variables.items():
-            if isinstance(value, int):
-                self._balance_variables[key] = value / 100
+            if isinstance(value, (int, float)):
+                # pylint: disable=line-too-long
+                self._balance_variables[key] = value / 100  # type: ignore[literal-required]
 
         self.last_balance_update = datetime.utcnow()
 
     @property
-    def account_number(self):
+    def account_number(self) -> Optional[str]:
         """
         Returns:
             str: the account's account number
@@ -79,24 +146,26 @@ class Account:
         return self.json.get("account_number")
 
     @property
-    def balance(self):
+    def balance(self) -> Optional[float]:
         """
         Returns:
             float: the currently available balance of the account
         """
-        return self._get_balance_property("balance")
+        return self._get_balance_property("balance")  # type: ignore[return-value]
 
     @property
-    def balance_including_flexible_savings(self):
+    def balance_including_flexible_savings(self) -> Optional[float]:
         """
         Returns:
             float: the currently available balance of the account, including flexible
              savings pots
         """
-        return self._get_balance_property("balance_including_flexible_savings")
+        return self._get_balance_property(
+            "balance_including_flexible_savings"
+        )  # type: ignore[return-value]
 
     @property
-    def created_datetime(self):
+    def created_datetime(self) -> Optional[datetime]:
         """
         Returns:
             datetime: when the account was created
@@ -107,7 +176,7 @@ class Account:
         return datetime.strptime(self.json["created"], DATETIME_FORMAT)
 
     @property
-    def description(self):
+    def description(self) -> Optional[str]:
         """
         Returns:
             str: the description of the account
@@ -115,15 +184,15 @@ class Account:
         return self.json.get("description")
 
     @property
-    def id(self):
+    def id(self) -> Optional[str]:
         """
         Returns:
             str: the account's UUID
         """
-        return self.json.get("id")
+        return self.json["id"]
 
     @property
-    def sort_code(self):
+    def sort_code(self) -> Optional[str]:
         """
         Returns:
             str: the account's sort code
@@ -131,33 +200,33 @@ class Account:
         return self.json.get("sort_code")
 
     @property
-    def spend_today(self):
+    def spend_today(self) -> Optional[float]:
         """
         Returns:
             float: the amount spent from this account today (considered from approx
              4am onwards)
         """
-        return self._get_balance_property("spend_today")
+        return self._get_balance_property("spend_today")  # type: ignore[return-value]
 
     @property
-    def total_balance(self):
+    def total_balance(self) -> Optional[str]:
         """
         Returns:
             str: the sum of the currently available balance of the account and the
              combined total of all the user’s pots
         """
-        return self._get_balance_property("total_balance")
+        return self._get_balance_property("total_balance")  # type: ignore[return-value]
 
 
 # pylint: disable=too-many-public-methods
 class Pot:
     """Read-only class for Monzo pots"""
 
-    def __init__(self, json):
+    def __init__(self, json: _MonzoPotInfo):
         self.json = json
 
     @property
-    def available_for_bills(self):
+    def available_for_bills(self) -> Optional[bool]:
         """
         Returns:
             bool: if the pot can be used directly for bills
@@ -165,15 +234,15 @@ class Pot:
         return self.json.get("available_for_bills")
 
     @property
-    def balance(self):
+    def balance(self) -> float:
         """
         Returns:
             float: the pot's balance in GBP
         """
-        return self.json.get("balance", 0) / 100
+        return float(self.json.get("balance", 0)) / 100
 
     @property
-    def charity_id(self):
+    def charity_id(self) -> Optional[str]:
         """
         Returns:
             str: not sure!
@@ -181,7 +250,7 @@ class Pot:
         return self.json.get("charity_id")
 
     @property
-    def cover_image_url(self):
+    def cover_image_url(self) -> Optional[str]:
         """
         Returns:
             str: URL for the cover image
@@ -189,7 +258,7 @@ class Pot:
         return self.json.get("cover_image_url")
 
     @property
-    def created_datetime(self):
+    def created_datetime(self) -> Optional[datetime]:
         """
         Returns:
             datetime: when the pot was created
@@ -200,7 +269,7 @@ class Pot:
         return datetime.strptime(self.json["created"], DATETIME_FORMAT)
 
     @property
-    def currency(self):
+    def currency(self) -> Optional[str]:
         """
         Returns:
             str: the currency of the pot
@@ -208,7 +277,7 @@ class Pot:
         return self.json.get("currency")
 
     @property
-    def current_account_id(self):
+    def current_account_id(self) -> Optional[str]:
         """
         Returns:
             str: the UUID of the parent account
@@ -216,7 +285,7 @@ class Pot:
         return self.json.get("current_account_id")
 
     @property
-    def deleted(self):
+    def deleted(self) -> Optional[bool]:
         """
         Returns:
             bool: has the pot been deleted
@@ -224,7 +293,7 @@ class Pot:
         return self.json.get("deleted")
 
     @property
-    def goal_amount(self):
+    def goal_amount(self) -> Optional[float]:
         """
         Returns:
             float: the user-set goal amount for the pot
@@ -235,7 +304,7 @@ class Pot:
         return goal_amount / 100
 
     @property
-    def has_virtual_cards(self):
+    def has_virtual_cards(self) -> Optional[bool]:
         """
         Returns:
             bool: if the pot has virtual cards attached
@@ -243,15 +312,15 @@ class Pot:
         return self.json.get("has_virtual_cards")
 
     @property
-    def id(self):
+    def id(self) -> str:
         """
         Returns:
             str: the pot's UUID
         """
-        return self.json.get("id")
+        return self.json["id"]
 
     @property
-    def is_tax_pot(self):
+    def is_tax_pot(self) -> Optional[bool]:
         """
         Returns:
             bool: is the pot taxed? I'm not sure
@@ -259,7 +328,7 @@ class Pot:
         return self.json.get("is_tax_pot")
 
     @property
-    def isa_wrapper(self):
+    def isa_wrapper(self) -> Optional[str]:
         """
         Returns:
             str: is the pot ISA-wrapped?
@@ -267,7 +336,7 @@ class Pot:
         return self.json.get("isa_wrapper")
 
     @property
-    def locked(self):
+    def locked(self) -> Optional[bool]:
         """
         Returns:
             bool: is the pot locked
@@ -275,15 +344,15 @@ class Pot:
         return self.json.get("locked")
 
     @property
-    def name(self):
+    def name(self) -> str:
         """
         Returns:
             str: the name of the pot
         """
-        return self.json.get("name")
+        return self.json["name"]
 
     @property
-    def product_id(self):
+    def product_id(self) -> Optional[str]:
         """
         Returns:
             str: the ID of the product applied to the pot (e.g. savings)
@@ -291,7 +360,7 @@ class Pot:
         return self.json.get("product_id")
 
     @property
-    def round_up(self):
+    def round_up(self) -> Optional[bool]:
         """
         Returns:
             bool: is the pot where all round ups go
@@ -299,15 +368,15 @@ class Pot:
         return self.json.get("round_up")
 
     @property
-    def round_up_multiplier(self):
+    def round_up_multiplier(self) -> Optional[float]:
         """
         Returns:
-            int: the multiplier applied to the pot's round-ups
+            float: the multiplier applied to the pot's round-ups
         """
         return self.json.get("round_up_multiplier")
 
     @property
-    def style(self):
+    def style(self) -> Optional[str]:
         """
         Returns:
             str: the pot background image
@@ -315,7 +384,7 @@ class Pot:
         return self.json.get("style")
 
     @property
-    def type(self):
+    def type(self) -> Optional[str]:
         """
         Returns:
             str: the type of pot (e.g. flex saver)
@@ -323,7 +392,7 @@ class Pot:
         return self.json.get("type")
 
     @property
-    def updated_datetime(self):
+    def updated_datetime(self) -> Optional[datetime]:
         """
         Returns:
             datetime: when the pot was updated last
@@ -333,7 +402,7 @@ class Pot:
 
         return datetime.strptime(self.json["updated"], DATETIME_FORMAT)
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"{self.name} | {self.id}"
 
 
@@ -347,31 +416,35 @@ class MonzoClient(OauthClient):
     def __init__(
         self,
         *,
-        client_id,
-        client_secret,
-        redirect_uri="http://0.0.0.0:5001/get_auth_code",
-        access_token_expiry_threshold=60,
-        log_requests=False,
-        creds_cache_path=None,
+        client_id: str,
+        client_secret: str,
+        redirect_uri: str = "http://0.0.0.0:5001/get_auth_code",
+        access_token_expiry_threshold: int = 60,
+        log_requests: bool = False,
+        creds_cache_path: Optional[str] = None,
     ):
         super().__init__(
             client_id=client_id,
             client_secret=client_secret,
+            base_url=self.BASE_URL,
+            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
             redirect_uri=redirect_uri,
             access_token_expiry_threshold=access_token_expiry_threshold,
             log_requests=log_requests,
-            creds_cache_path=creds_cache_path,
+            creds_cache_path=creds_cache_path or self.CREDS_FILE_PATH,
         )
 
-        self._current_account = None
+        self._current_account: Optional[Account] = None
 
-    def deposit_into_pot(self, pot, amount_pence, dedupe_id=None):
+    def deposit_into_pot(
+        self, pot: Pot, amount_pence: int, dedupe_id: Optional[str] = None
+    ) -> None:
         """Move money from an account owned by the currently authorised user into one
          of their pots
 
         Args:
             pot (Pot): the target pot
-            amount_pence (int): the amount of money to depoist, in pence
+            amount_pence (int): the amount of money to deposit, in pence
             dedupe_id (str): unique string used to de-duplicate deposits. Will be
              created if not provided
         """
@@ -389,7 +462,9 @@ class MonzoClient(OauthClient):
         )
         res.raise_for_status()
 
-    def list_accounts(self, ignore_closed=True, account_type=None):
+    def list_accounts(
+        self, ignore_closed: bool = True, account_type: Optional[str] = None
+    ) -> Generator[Account, None, None]:
         """Gets a list of the user's accounts
 
         Args:
@@ -410,7 +485,7 @@ class MonzoClient(OauthClient):
                 continue
             yield Account(account, self)
 
-    def list_pots(self, ignore_deleted=True):
+    def list_pots(self, ignore_deleted: bool = True) -> Generator[Pot, None, None]:
         """Gets a list of the user's pots
 
         Args:
@@ -429,7 +504,7 @@ class MonzoClient(OauthClient):
                 continue
             yield Pot(pot)
 
-    def get_pot_by_id(self, pot_id):
+    def get_pot_by_id(self, pot_id: str) -> Optional[Pot]:
         """Get a pot from its ID
 
         Args:
@@ -444,7 +519,9 @@ class MonzoClient(OauthClient):
 
         return None
 
-    def get_pot_by_name(self, pot_name, exact_match=False):
+    def get_pot_by_name(
+        self, pot_name: str, exact_match: bool = False
+    ) -> Optional[Pot]:
         """Get a pot from its name
 
         Args:
@@ -459,16 +536,14 @@ class MonzoClient(OauthClient):
             pot_name = cleanse_string(pot_name)
 
         for pot in self.list_pots():
-            found_name = (
-                cleanse_string(pot.name).lower() if not exact_match else pot_name
-            )
+            found_name = pot_name if exact_match else cleanse_string(pot.name).lower()
             if found_name.lower() == pot_name.lower():
                 return pot
 
         return None
 
     @property
-    def access_token_has_expired(self):
+    def access_token_has_expired(self) -> bool:
         """Custom expiry check for Monzo client as JWT doesn't seem to include expiry
          time. Any errors/missing credentials result in a default value of True
 
@@ -477,7 +552,7 @@ class MonzoClient(OauthClient):
         """
         self._load_local_credentials()
 
-        if not (access_token := self._credentials.get("access_token")):
+        if not (access_token := self._credentials.get("access_token", False)):
             return True
 
         res = get(
@@ -488,7 +563,7 @@ class MonzoClient(OauthClient):
         return res.json().get("authenticated", False) is False
 
     @property
-    def current_account(self):
+    def current_account(self) -> Account:
         """Get the main account for the Monzo user. We assume there'll only be one
          main account per user
 
